#ifndef WINDOWS_INCLUDED
#define WINDOWS_INCLUDED
#include "patchlevel.h"

#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include "effects.h"
#include "video.h"

#define NPAGES 2

typedef
    struct {
	int	    xloc, yloc;
        unsigned long int Width, Height;
        unsigned int maxcolor;
        unsigned int maxcells;
        int         cscalemax;
        int         pixwid;
        int         pixht;
	int	    xcmult;
	int	    ycmult;
	int	    xoffs;
	int	    yoffs;
        Display     *theDisp;
        Colormap    theCmap;
        GC          xorGC;
        GC          clearbitsGC;
        GC          overlayGC;
        GC          clearGC;
        GC          copyGC;
        GC          createGC;
        GC          tileGC;
        GC          drawGC;
        GC          edgeGC;
        int         theScreen;
        int         theDepth;
	unsigned long pixvals[MAXCOLOR];
	unsigned long planes;
	unsigned long fcol, bcol;
        Visual      *theVisual;
        Window      RootW;
        Window      theWindow;
	Pixmap	    thePix;
	Pixmap	    winPix;
	Pixmap	    page[NPAGES];
	byte	    current_mode;
    } WINDOW;

typedef struct CMap_struct { /* Sparse colormap */
    XColor clist[MAXCOLOR];
    int cindx[MAXCOLOR];
    int ncolors;
    int last_upd;
} *CMap;

extern WINDOW MainWindow;
extern CMap current_cmap;
extern BOOLEAN edge_mode;
extern BOOLEAN transparent_mode;
extern BOOLEAN is_transparent[MAXCOLOR];
extern unsigned long int pixarr[MAXCOLOR];

#define index_to_pixel(i) (pixarr[i])

#define MAXPIXEL 65535
#define NUMFONTBUFS 32
typedef struct {
    unsigned charcnt,width,height,firstch,style,bytes_per_char,
	cgap,sgap,def_cgap,def_sgap;
    struct {
	int firstx, lastx, width;
    } *cwidths;
    XImage *img;
    BOOLEAN active;
} MyFont;
extern MyFont fonts[];
extern MyFont *curfont;

#define FSTYLE_BOLD_UP			1
#define FSTYLE_BOLD_RIGHT		2
#define FSTYLE_SHADOW_UP_RIGHT		3
#define FSTYLE_SHADOW_UP_LEFT		4
#define FSTYLE_SHADOW_2UP_2RIGHT	5
#define FSTYLE_SHADOW_2UP_2LEFT		6

#define TO_WINDOW TRUE
#define TO_DRAWING FALSE
#ifdef CLIP_ALL_BUT_CFADE
#define put_preg(r,x,y,e,s,d,em) put_reg(r,x,y,e,s,FALSE,TRUE,d,em)
#define put_creg put_preg
#define put_transient(r,x,y,e,s,d,em) put_reg(r,x,y,e,s,TRUE,TRUE,d,em)
#else
#define put_preg(r,x,y,e,s,d,em) put_reg(r,x,y,e,s,FALSE,TRUE,d,em)
#define put_creg(r,x,y,e,s,d,em) put_reg(r,x,y,e,s,FALSE,FALSE,d,em)
#define put_transient(r,x,y,e,s,d,em) put_reg(r,x,y,e,s,TRUE,FALSE,d,em)
#endif
#define save_bg(e) TRUE
#define screen_x(x) ((((x)*MainWindow.xcmult)+MainWindow.xoffs)*MainWindow.pixwid)
#define screen_y(y,ht) (MainWindow.Height-((((y)*MainWindow.ycmult)+MainWindow.yoffs)*MainWindow.pixht)-(ht))

#include "registers.h"
BOOLEAN convert_gif_for_screen();
BOOLEAN convert_img_for_screen();
BOOLEAN create_font();
BOOLEAN set_font_default();
BOOLEAN create_preg();
BOOLEAN cycle_colors();
BOOLEAN get_reg();
BOOLEAN init_window();
BOOLEAN pan_preg();
BOOLEAN put_partial();
BOOLEAN put_reg();
BOOLEAN set_clipping_region();
BOOLEAN set_disp_page();
BOOLEAN set_drawing_reg();
BOOLEAN setup_creg_for_display();
BOOLEAN setup_preg_for_display();
BOOLEAN setup_for_trans_mode();
BOOLEAN spread_colors();
BOOLEAN start_window();
BOOLEAN tile_window();
CMap convert_clist_to_cmap();
int evt_handler();
int get_keystroke();
void activate_current_cmap();
void beep();
void cleanup();
void clear_creg();
void clear_preg();
void clear_window();
void copy_cmap();
void copy_pix();
void dispose_font();
void drawbox();
void drawellipse();
void drawline();
void drawpoint();
void flush_windows();
void put_char();
void redraw_screen();
void restore_bg();
void set_current_cmap_entry();
void set_current_drawcolor();
void set_current_drawcolor2();
void set_edge_color();
void update_current_cmap();
void synchronize_display();

#endif
